package revolution.test;

import java.io.File;
import java.time.Instant;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import org.hipparchus.util.FastMath;
import org.orekit.bodies.GeodeticPoint;
import org.orekit.bodies.OneAxisEllipsoid;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.frames.TopocentricFrame;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.analytical.tle.TLE;
import org.orekit.propagation.analytical.tle.TLEPropagator;
import org.orekit.propagation.events.ElevationDetector;
import org.orekit.propagation.events.EventDetector;
import org.orekit.propagation.events.LatitudeCrossingDetector;
import org.orekit.propagation.events.handlers.EventHandler;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScale;
import org.orekit.time.TimeScalesFactory;
import org.orekit.utils.Constants;
import org.orekit.utils.IERSConventions;

/**
 *
 * @author eduan
 */
public class Main {

    private static final double MAX_CHECK = 60.0;
    private static final double THRESHOLD = 0.001;
    private static final double ELEVATION = FastMath.toRadians(5.0);
    private static final double STATION_LAT = FastMath.toRadians(29.975995);
    private static final double STATION_LON = FastMath.toRadians(31.130753);
    private static final double STATION_ALT = 210;
    private static final DateTimeFormatter DATE_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss'Z'");

    private final OneAxisEllipsoid earth;
    private final TopocentricFrame stationFrame;
    private final TimeScale utcScale;

    public Main() throws Exception {
        File orekitData = new File("./"); // copy orekit-data-master.zip to current directory before running
        DataProvidersManager manager = DataProvidersManager.getInstance();
        manager.addProvider(new DirectoryCrawler(orekitData));

        this.utcScale = TimeScalesFactory.getUTC();
        GeodeticPoint station = new GeodeticPoint(STATION_LAT, STATION_LON, STATION_ALT);
        Frame earthFrame = FramesFactory.getITRF(IERSConventions.IERS_2010, true);
        this.earth = new OneAxisEllipsoid(Constants.WGS84_EARTH_EQUATORIAL_RADIUS, Constants.WGS84_EARTH_FLATTENING, earthFrame);
        this.stationFrame = new TopocentricFrame(this.earth, station, "station");
    }

    private void printContacts(TLE tle, Date from, Date to) throws Exception {
        AbsoluteDate startAbsolute = new AbsoluteDate(from, this.utcScale);
        AbsoluteDate endAbsolute = new AbsoluteDate(to, this.utcScale);

        long fromSSE = startAbsolute.toDate(this.utcScale).getTime();
        Date[] lastStart = new Date[1];
        int[] revolutionNumber = new int[1];
        int[] lastRevolution = new int[1];
        TLEPropagator propagator = TLEPropagator.selectExtrapolator(tle);

        lastRevolution[0] = tle.getRevolutionNumberAtEpoch();

        EventDetector stationDetector = new ElevationDetector(MAX_CHECK, THRESHOLD, this.stationFrame).withConstantElevation(ELEVATION)
                .withHandler((SpacecraftState state, ElevationDetector d, boolean increasing) -> {
                    Date date = state.getDate().toDate(this.utcScale);
                    if (date.getTime() >= fromSSE) {
                        if (increasing) {
                            lastStart[0] = date;
                            revolutionNumber[0] = lastRevolution[0];
                        } else {
                            if (lastStart[0] != null) {
                                String start = ZonedDateTime.ofInstant(Instant.ofEpochMilli(lastStart[0].getTime()), ZoneOffset.UTC).format(DATE_FORMATTER);
                                String end = ZonedDateTime.ofInstant(Instant.ofEpochMilli(date.getTime()), ZoneOffset.UTC).format(DATE_FORMATTER);
                                System.out.println(String.format("   Contact with revolution %d from %s to %s", revolutionNumber[0], start, end));
                                lastStart[0] = null;
                            }
                        }
                    }

                    return EventHandler.Action.CONTINUE;
                });

        EventDetector equatorCrossingDetector = new LatitudeCrossingDetector(MAX_CHECK, THRESHOLD, this.earth, STATION_LAT)
                .withHandler((SpacecraftState state, LatitudeCrossingDetector detector, boolean increasing) -> {
                    if (increasing) {
                        lastRevolution[0] = lastRevolution[0] + 1;
                    }
                    return EventHandler.Action.CONTINUE;
                });

        propagator.addEventDetector(equatorCrossingDetector);
        propagator.addEventDetector(stationDetector);

        propagator.propagate(endAbsolute);
    }

    public void run() throws Exception {
        Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        cal.set(2019, 10, 27, 21, 40, 0);

        Date start = cal.getTime();
        Date end = Date.from(Instant.ofEpochMilli(start.getTime()).plus(24, ChronoUnit.HOURS));

        TLE tle0 = new TLE("1 38771U 12049A   19331.44643519  .00000016  00000-0  27590-4 0  9999", "2 38771  98.7226  28.5416 0000787 163.2861  45.2811 14.21479650373178");
        TLE tle1 = new TLE("1 38771U 12049A   19331.89619213  .00000006  00000-0  22710-4 0  9990", "2 38771  98.7225  28.9859 0000807 162.8377 185.9845 14.21479658373247");

        String tle0Date = ZonedDateTime.ofInstant(Instant.ofEpochMilli(tle0.getDate().toDate(this.utcScale).getTime()), ZoneOffset.UTC).format(DATE_FORMATTER);
        String tle1Date = ZonedDateTime.ofInstant(Instant.ofEpochMilli(tle1.getDate().toDate(this.utcScale).getTime()), ZoneOffset.UTC).format(DATE_FORMATTER);

        System.out.println("\nContacts via first TLE (with date " + tle0Date + "):");
        this.printContacts(tle0, start, end);

        System.out.println("\nContacts via next TLE (with date " + tle1Date + "):");
        this.printContacts(tle1, start, end);
    }

    public static void main(String[] args) throws Exception {
        Main main = new Main();
        main.run();
    }

}
